'' =================================================================================================
''
''   File....... jm_pwm8.spin
''   Purpose.... 1-to-8 channel PWM driver
''               -- uses contiguous pins
''               -- pwm frequency is ~1.85kHz with 80MHz system clock
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2017 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon@jonmcphalen.com
''   Started....
''   Updated.... 11 SEP 2017
''
'' =================================================================================================


con { io pins }                                                  
                                                                 
  RX1 = 31  { I }                                               ' serial / programming
  TX1 = 30  { O }                                            
                                                             
  SDA = 29  { I/O }                                             ' i2c / eeprom
  SCL = 28  { I/O } 


var

  long  cog
  long  levels[2]                                               ' long aligned for fast access by pwm cog


pub null

  ' This is not a top-level object


pub start(count, base)

'' Start PWM8 pwm driver on contiguous pins; uses one cog
'' -- count is number of pwm outputs (1 to 8)
'' -- base is the first pin (lsb) to use

  stop                                                          ' stop if running

  levels.byte[0] := 1 #> count <#8
  levels.byte[1] := base
                                                             
  cog := cognew(@pwm8, @levels) + 1                             ' start the pwm cog

  if (cog)                                                      ' if it started
    repeat until (levels == 0)                                  '  let it get setup
                                                                   
  return cog                                                     
                
  
pub stop

'' Stop PWM cog if running

  if (cog)
    cogstop(cog - 1)
    cog := 0

  longfill(@levels, 0, 2)


pub set(ch, level)

'' Set channel to level (0..255)
                                             
  if ((ch => 0) and (ch =< 7)) 
    levels.byte[ch] := 0 #> level <# 255


pub set_all(level)

'' Set all channels to level

  bytefill(@levels, 0 #> level <# 255, 8)


pub high(ch)

'' Set channel to 100%

  if ((ch => 0) and (ch =< 7))
    levels.byte[ch] := 255


pub low(ch)

'' Set channel to 0%

  if ((ch => 0) and (ch =< 7))
    levels.byte[ch] := 0


pub toggle(ch)

'' Toggle state of channel
'' -- make 100% if below 50%
'' -- make   0% if equal or above 50%

  if ((ch => 0) and (ch =< 7))
    if (levels.byte[ch] < 128)
      levels.byte[ch] := 255
    else
      levels.byte[ch] := 0


pub read(ch)

'' Read current channel level (0..255)
'' -- returns -1 if bad channel specified

  if ((ch => 0) and (ch =< 7))
    return levels.byte[ch]
  else
    return -1 


pub gamma8(level)

'' Convert linear level (0..255) to Gamma curve
'' -- improves apparent linearity

  if ((level => 0) and (level =< 255))
    return GammaTable[level] 


pub address

'' Expose hub address of PWM levels
'' -- use for direct control of levels

  return @levels


dat { gamma table }

  GammaTable    byte      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0
                byte      0,   0,   0,   0,   0,   0,   0,   0,   1,   1,   1,   1,   1,   1,   1,   1
                byte      1,   1,   1,   1,   2,   2,   2,   2,   2,   2,   2,   2,   3,   3,   3,   3
                byte      3,   3,   4,   4,   4,   4,   5,   5,   5,   5,   5,   6,   6,   6,   6,   7
                byte      7,   7,   8,   8,   8,   9,   9,   9,  10,  10,  10,  11,  11,  11,  12,  12
                byte     13,  13,  13,  14,  14,  15,  15,  16,  16,  17,  17,  18,  18,  19,  19,  20
                byte     20,  21,  21,  22,  22,  23,  24,  24,  25,  25,  26,  27,  27,  28,  29,  29
                byte     30,  31,  31,  32,  33,  34,  34,  35,  36,  37,  38,  38,  39,  40,  41,  42
                byte     42,  43,  44,  45,  46,  47,  48,  49,  50,  51,  52,  53,  54,  55,  56,  57
                byte     58,  59,  60,  61,  62,  63,  64,  65,  66,  68,  69,  70,  71,  72,  73,  75
                byte     76,  77,  78,  80,  81,  82,  84,  85,  86,  88,  89,  90,  92,  93,  94,  96
                byte     97,  99, 100, 102, 103, 105, 106, 108, 109, 111, 112, 114, 115, 117, 119, 120
                byte    122, 124, 125, 127, 129, 130, 132, 134, 136, 137, 139, 141, 143, 145, 146, 148
                byte    150, 152, 154, 156, 158, 160, 162, 164, 166, 168, 170, 172, 174, 176, 178, 180
                byte    182, 184, 186, 188, 191, 193, 195, 197, 199, 202, 204, 206, 209, 211, 213, 215
                byte    218, 220, 223, 225, 227, 230, 232, 235, 237, 240, 242, 245, 247, 250, 252, 255


dat { pwm driver }

' Influenced by work from Beau Schwabe

                        org     0

pwm8                    movd    :wr_mask, #chmasks              ' point to cog mask table
                        mov     mastermask, #0                  ' clear master mask

                        rdlong  t1, par                         ' get ch count and base pin                   
                        mov     t2, t1                          ' copy for base pin
                        and     t1, #$0F                        ' isolate count
                        shr     t2, #8                          ' isolate base pin
                        and     t2, #$1F
                        mov     t3, #1                          ' create base pin mask
                        shr     t3, t2

:wr_mask                mov     0-0, t3                         ' save channel mask
                        or      mastermask, t3                  ' build master mask
                        shl     t3, #1                          ' update ch mask for next pin
                        add     :wr_mask, INC_DEST              ' point to next mask in cog table
                        djnz    t1, #:wr_mask                   ' build masks for defined pins

                        mov     dira, mastermask                ' set pins used to output      

                        mov     t1, #0                          ' alert fg that cog is ready
                        wrlong  t1, par


pwm_main                mov     t1, par                         ' read current levels
                        rdlong  chlevels + 0, t1
                        add     t1, #4
                        nop
                        rdlong  chlevels + 1, t1

                        mov     phase, #0                       ' reset cycle
                        mov     outsmask, mastermask            ' enable defined pins

                        
ch00                    mov     t1, chlevels + 0                ' get ch level
                        and     t1, #$FF                wz      ' isolate, check for 0
                        cmp     t1, phase               wc      ' finished?
  if_z_or_c             andn    outsmask, chmasks + 00          ' off if 0 or finished  

  
ch01                    mov     t1, chlevels + 0
                        shr     t1, #8              
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 01         

  
ch02                    mov     t1, chlevels + 0
                        shr     t1, #16             
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 02


ch03                    mov     t1, chlevels + 0
                        shr     t1, #24             
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 03


ch04                    mov     t1, chlevels + 1
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 04

   
ch05                    mov     t1, chlevels + 1
                        shr     t1, #8             
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 05


ch06                    mov     t1, chlevels + 1
                        shr     t1, #16             
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 06


ch07                    mov     t1, chlevels + 1
                        shr     t1, #24           
                        and     t1, #$FF                wz 
                        cmp     t1, phase               wc 
  if_z_or_c             andn    outsmask, chmasks + 07
  
  
update_outs             mov     outa, outsmask                  ' update outputs

                        add     phase, #1                       ' advance phase
                        and     phase, #$FF             wz      ' reset at 256
  if_nz                 jmp     #ch00                           ' back to ch00 if still running
                      
                        jmp     #pwm_main                       ' read levels and start again


' -------------------------------------------------------------------------------------------------

INC_DEST                long    1 << 9                          ' to increment D field

chmasks                 long    0[8]                            ' channel masks

mastermask              res     1                               ' master mask for all channels
chlevels                res     2                               ' channel levels
outsmask                res     1                               ' working outputs mask
phase                   res     1                               ' current phase in pwm cycle                            

t1                      res     1                               ' work vars
t2                      res     1
t3                      res     1

                        fit     496

                                  
dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      